"""Documentation about segmentmytif."""

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Christiaan Meijer"
__email__ = "c.meijer@esciencecenter.nl"
__version__ = "0.3.0"
